/*
 */
/*
 * dgram-socket create <local>                        => <sockid>
 * dgram-socket close <sockid>
 * dgram-socket transmit <sockid> <data> <remote>
 * dgram-socket on-receive <sockid> [<script>]
 *    calls, effectively,  eval <script> [list <data> <remote-addr> <socket>]
 *    if script not supplied, cancel
 */

#include "tables.h"
#include "hbytes.h"

typedef struct DgramSocket {
  int ix; /* first ! */
  int fd, script_llength;
  Tcl_Interp *ip;
  Tcl_Obj *script;
  void *addr_buf, *msg_buf;
  int addr_buflen, msg_buflen;
} DgramSocket;

IdDataTable dgram_socks= { "dgramsock" };

int do_dgram_socket_create(ClientData cd, Tcl_Interp *ip,
			   SockAddr_Value local, void **sock_r) {
  int fd, al, r;
  DgramSocket *sock;
  const struct sockaddr *sa;

  sa= sockaddr_addr(&local);
  al= sockaddr_len(&local);

  fd= socket(sa->sa_family, SOCK_DGRAM, 0);
  if (fd<0) return posixerr(ip,errno,"socket");
  r= bind(fd, sa, al);  if (r) return newfdposixerr(ip,fd,"bind");
  r= setnonblock(fd, 1);  if (r) return newfdposixerr(ip,fd,"setnonblock");

  sock= TALLOC(sizeof(DgramSocket));
  sock->ix= -1;
  sock->fd= fd;
  sock->script= 0;
  sock->addr_buflen= al+1;
  sock->addr_buf= TALLOC(sock->addr_buflen);
  sock->msg_buflen= 0;
  sock->msg_buf= 0;

  *sock_r= sock;
  return TCL_OK;
}

int do_dgram_socket_transmit(ClientData cd, Tcl_Interp *ip,
			     void *sock_v, HBytes_Value data,
			     SockAddr_Value remote) {
  DgramSocket *sock= sock_v;
  int l, r;

  r= sendto(sock->fd,
	    hbytes_data(&data), l=hbytes_len(&data),
	    0,
	    sockaddr_addr(&remote), sockaddr_len(&remote));
  if (r==-1) return posixerr(ip,errno,"sendto");
  else if (r!=l) return staticerr(ip,"sendto gave wrong answer",0);
  return TCL_OK;
}

static void cancel(DgramSocket *sock) {
  if (sock->script) {
    Tcl_DeleteFileHandler(sock->fd);
    Tcl_DecrRefCount(sock->script);
    sock->script= 0;
  }
}

static void recv_call(ClientData sock_cd, int mask) {
  DgramSocket *sock= (void*)sock_cd;
  Tcl_Interp *ip= sock->ip;
  int sz, rc, i, peek;
  HBytes_Value message_val;
  SockAddr_Value peer_val;
  Tcl_Obj *args[3], *invoke;
  struct msghdr mh;
  struct iovec iov;

  hbytes_empty(&message_val);
  sockaddr_clear(&peer_val);
  invoke=0; for (i=0; i<3; i++) args[i]=0;

  mh.msg_iov= &iov;
  mh.msg_iovlen= 1;
  mh.msg_control= 0;
  mh.msg_controllen= 0;
  mh.msg_flags= 0;

  peek= MSG_PEEK;
  
  for (;;) {
    mh.msg_name= sock->addr_buf;
    mh.msg_namelen= sock->addr_buflen;

    iov.iov_base= sock->msg_buf;
    iov.iov_len= sock->msg_buflen;

    sz= recvmsg(sock->fd, &mh, peek);
    if (sz==-1) { rc=0; goto x_rc; }

    assert(mh.msg_namelen < sock->addr_buflen);

    if (!(mh.msg_flags & MSG_TRUNC)) {
      if (!peek) break;
      peek= 0;
      continue;
    }

    TFREE(sock->msg_buf);
    sock->msg_buflen *= 2;
    sock->msg_buflen += 100;
    sock->msg_buf= TALLOC(sock->msg_buflen);
  }

  hbytes_array(&message_val, iov.iov_base, sz);
  sockaddr_create(&peer_val, mh.msg_name, mh.msg_namelen);

  args[0]= ret_hb(ip, message_val);  hbytes_empty(&message_val);
  args[1]= ret_sockaddr(ip, peer_val);  sockaddr_clear(&peer_val);
  args[2]= ret_iddata(ip, sock, &dgram_socks);
  for (i=0; i<3; i++) Tcl_IncrRefCount(args[i]);

  invoke= Tcl_DuplicateObj(sock->script);
  Tcl_IncrRefCount(invoke);

  rc= Tcl_ListObjReplace(ip,invoke,sock->script_llength,0,3,args);
  for (i=0; i<3; i++) { Tcl_DecrRefCount(args[i]); args[i]= 0; }
  if (rc) goto x_rc;

  rc= Tcl_EvalObjEx(ip,invoke,TCL_EVAL_GLOBAL|TCL_EVAL_DIRECT);

x_rc:
  if (invoke) Tcl_DecrRefCount(invoke);

  if (rc)
    Tcl_BackgroundError(ip);
}

int do_dgram_socket_on_receive(ClientData cd, Tcl_Interp *ip,
			       void *sock_v, Tcl_Obj *script) {
  DgramSocket *sock= sock_v;
  int rc;
  
  if (script) {
    rc= Tcl_ListObjLength(ip, script, &sock->script_llength);
    if (rc) return rc;
  }
  
  cancel(sock);
  if (script) {
    Tcl_IncrRefCount(script);
    sock->script= script;
    sock->ip= ip;
  }
  Tcl_CreateFileHandler(sock->fd, TCL_READABLE, recv_call, sock);
  return TCL_OK;
}

int do_dgram_socket_close(ClientData cd, Tcl_Interp *ip, void *sock_v) {
  DgramSocket *sock= sock_v;
  int sockix;
  cancel(sock);
  close(sock->fd); /* nothing useful to be done with errors */
  sockix= sock->ix;
  TFREE(sock->addr_buf);
  TFREE(sock->msg_buf);
  TFREE(sock);
  dgram_socks.a[sockix]= 0;
  return TCL_OK;
}
