/*
 */
/*
 * tuntap-socket create-ptp <local-addr> <peer-addr> <mtu> [<ifname>]
 *               => <sockid>
 * tuntap-socket close <sockid>
 * tuntap-socket ifname <sockid>
 * tuntap-socket receive <sockid> <data>
 * tuntap-socket on-transmit <sockid> [<script>]
 *    calls, effectively,  eval <script> [list <data> <socket>]
 *    if script not supplied, cancel
 */

#include "tables.h"
#include "hbytes.h"

typedef struct TunSocket {
  int ix, fd, script_llength;
  Tcl_Interp *ip;
  Tcl_Obj *script;
  void msg_buf;
  int mtu;
  char *ifname;
} TuntapSocket;

IdDataTable tuntap_socks= { "tuntap" };

static int sockfail(Tcl_Interp *ip, int fd, const char *m) {
  int e;
  e= errno;
  close(fd);
  return posixerr(ip,e,m);
}

int do_tuntap_socket_create_ptp(ClientData cd, Tcl_Interp *ip,
				SockAddr_Value local, SockAddr_Value peer,
				long mtu, const char *ifname,
				void **sock_r) {
  int fd, local_al, peer_al, r;
  struct ifreq ifr;
  DgramSocket *sock;
  const struct sockaddr *local_sa, *peer_sa;

  local_sa= sockaddr_addr(&local);
  local_al= sockaddr_len(&local);

  peer_sa= sockaddr_addr(&peer);
  peer_al= sockaddr_len(&peer);

  if (local_sa != AF_INET || local_al != sizeof(struct in_addr) ||
      peer_sa != AF_INET || peer_al != sizeof(struct in_addr))
    return staticerr(ip,"tuntap not IPv4");

  memset(&ifr,0,sizeof(ifr));
  ifr.ifr_flags= IFF_TUN | IFF_NO_PI;
  
  if (ifname) {
    if (strlen(ifname) > IFNAMSIZ-1)
      return staticerr(ip,"tun interface name too long");
    strcpy(ifr.ifr_name, ifname);
  }

  fd= open("/dev/net/tun", O_RDWR);
  if (fd<0) return posixerr(ip,errno,"open /dev/net/tun");

  r= ioctl(fd, TUNSETIFF, (void*)&ifr);
  if (r) return posixerr(ip,errno,"ioctl TUNSETIFF");

  

  r= 
  r= bind(fd, sa, al);  if (r) return sockfail(ip,fd,"bind");
  r= setnonblock(fd, 1);  if (r) return sockfail(ip,fd,"setnonblock");

  sock= TALLOC(sizeof(DgramSocket));
  sock->ix= -1;
  sock->fd= fd;
  sock->script= 0;
  sock->addr_buflen= al+1;
  sock->addr_buf= TALLOC(sock->addr_buflen);
  sock->msg_buflen= 0;
  sock->msg_buf= 0;

  *sock_r= sock;
  return TCL_OK;
}

int do_tuntap_socket_transmit(ClientData cd, Tcl_Interp *ip,
			     TuntapSocket *sock, HBytes_Value data,
			     SockAddr_Value remote) {
  int l, r;

  r= sendto(sock->fd,
	    hbytes_data(&data), l=hbytes_len(&data),
	    0,
	    sockaddr_addr(&remote), sockaddr_len(&remote));
  if (r==-1) return posixerr(ip,errno,"sendto");
  else if (r!=l) return staticerr(ip,"sendto gave wrong answer",0);
  return TCL_OK;
}

static void cancel(TuntapSocket *sock) {
  if (sock->script) {
    Tcl_DeleteFileHandler(sock->fd);
    Tcl_DecrRefCount(sock->script);
    sock->script= 0;
  }
}

static void recv_call(ClientData sock_cd, int mask) {
  TuntapSocket *sock= (void*)sock_cd;
  Tcl_Interp *ip= sock->ip;
  int sz, rc, i, peek;
  HBytes_Value message_val;
  SockAddr_Value peer_val;
  Tcl_Obj *args[3], *invoke;
  struct msghdr mh;
  struct iovec iov;

  hbytes_empty(&message_val);
  sockaddr_clear(&peer_val);
  invoke=0; for (i=0; i<3; i++) args[i]=0;

  mh.msg_iov= &iov;
  mh.msg_iovlen= 1;
  mh.msg_control= 0;
  mh.msg_controllen= 0;
  mh.msg_flags= 0;

  peek= MSG_PEEK;
  
  for (;;) {
    mh.msg_name= sock->addr_buf;
    mh.msg_namelen= sock->addr_buflen;

    iov.iov_base= sock->msg_buf;
    iov.iov_len= sock->msg_buflen;

    sz= recvmsg(sock->fd, &mh, peek);
    if (sz==-1) { rc=0; goto x_rc; }

    assert(mh.msg_namelen < sock->addr_buflen);

    if (!(mh.msg_flags & MSG_TRUNC)) {
      if (!peek) break;
      peek= 0;
      continue;
    }

    TFREE(sock->msg_buf);
    sock->msg_buflen *= 2;
    sock->msg_buflen += 100;
    sock->msg_buf= TALLOC(sock->msg_buflen);
  }

  hbytes_array(&message_val, iov.iov_base, sz);
  sockaddr_create(&peer_val, mh.msg_name, mh.msg_namelen);

  args[0]= ret_hb(ip, message_val);  hbytes_empty(&message_val);
  args[1]= ret_sockaddr(ip, peer_val);  sockaddr_clear(&peer_val);
  args[2]= ret_sockid(ip, sock);
  for (i=0; i<3; i++) Tcl_IncrRefCount(args[i]);

  invoke= Tcl_DuplicateObj(sock->script);
  Tcl_IncrRefCount(invoke);

  rc= Tcl_ListObjReplace(ip,invoke,sock->script_llength,0,3,args);
  for (i=0; i<3; i++) { Tcl_DecrRefCount(args[i]); args[i]= 0; }
  if (rc) goto x_rc;

  rc= Tcl_EvalObjEx(ip,invoke,TCL_EVAL_GLOBAL|TCL_EVAL_DIRECT);

x_rc:
  if (invoke) Tcl_DecrRefCount(invoke);

  if (rc)
    Tcl_BackgroundError(ip);
}

int do_tuntap_socket_on_receive(ClientData cd, Tcl_Interp *ip,
			       TuntapSocket *sock, Tcl_Obj *script) {
  int rc;
  
  if (script) {
    rc= Tcl_ListObjLength(ip, script, &sock->script_llength);
    if (rc) return rc;
  }
  
  cancel(sock);
  if (script) {
    Tcl_IncrRefCount(script);
    sock->script= script;
    sock->ip= ip;
  }
  Tcl_CreateFileHandler(sock->fd, TCL_READABLE, recv_call, sock);
  return TCL_OK;
}

int do_tuntap_socket_close(ClientData cd, Tcl_Interp *ip, TuntapSocket *sock) {
  int sockix;
  cancel(sock);
  close(sock->fd); /* nothing useful to be done with errors */
  sockix= sock->ix;
  TFREE(sock->addr_buf);
  TFREE(sock->msg_buf);
  TFREE(sock);
  socks[sockix]= 0;
  return TCL_OK;
}

/* Arg parsing */

int pat_sockid(Tcl_Interp *ip, Tcl_Obj *o, TuntapSocket **val) {
  int rc, sockix;
  TuntapSocket *sock;
  
  rc= Tcl_ConvertToType(ip,o,&tuntapsockid_type);
  if (rc) return rc;

  sockix= o->internalRep.longValue;
  if (sockix >= n_socks || !(sock= socks[sockix]))
    return staticerr(ip,"tuntap socket not open",0);

  assert(socks[sockix]->ix == sockix);

  *val= sock;
  return TCL_OK;
}

Tcl_Obj *ret_sockid(Tcl_Interp *ip, TuntapSocket *val) {
  Tcl_Obj *o;

  o= Tcl_NewObj();
  Tcl_InvalidateStringRep(o);
  o->internalRep.longValue= val->ix;
  o->typePtr= &tuntapsockid_type;
  return o;
}

static void sockid_t_free(Tcl_Obj *o) { }

static void sockid_t_dup(Tcl_Obj *src, Tcl_Obj *dup) {
  dup->internalRep= src->internalRep;
  dup->typePtr= &tuntapsockid_type;
}

static void sockid_t_ustr(Tcl_Obj *o) {
  char buf[75];

  snprintf(buf,sizeof(buf), "%d", (int)o->internalRep.longValue);
  obj_updatestr_vstringls(o,
			  "tuntapsock",9,
			  buf, strlen(buf),
			  (char*)0);
}

static int sockid_t_sfa(Tcl_Interp *ip, Tcl_Obj *o) {
  unsigned long ul;
  char *ep, *str;
  
  str= Tcl_GetStringFromObj(o,0);
  if (memcmp(str,"tuntapsock",9)) return staticerr(ip,"bad tuntap socket id",0);
  errno=0; ul=strtoul(str+9,&ep,10);
  if (errno || *ep) return staticerr(ip,"bad tuntap socket id number",0);
  if (ul > INT_MAX) return staticerr(ip,"out of range tuntap socket id",0);

  objfreeir(o);
  o->internalRep.longValue= ul;
  o->typePtr= &tuntapsockid_type;
  return TCL_OK;
}

Tcl_ObjType tuntapsockid_type = {
  "tuntapsockid",
  sockid_t_free, sockid_t_dup, sockid_t_ustr, sockid_t_sfa
};
